from typing import Any
from pysome import SameState


class expect:
    def __init__(self, data: Any):
        self.data = data

    def to_be(self, other):
        SameState._start()  # noqa
        assert other == self.data
        SameState._end()  # noqa
        return self

    def not_to_be(self, other):
        SameState._start()  # noqa
        assert not other == self.data
        SameState._end()  # noqa
        return self



from typing import Callable, Union

from pysome import Some, AllOf
from pysome.SameState import default_name, SameState
from pysome.exceptions import InvalidArgument, SameOutsideExpect


class Same(Some):
    state_name = "Same"

    def __init__(self, *args: Union[type, Callable, Some], name=default_name):
        self.some = Some(*args)
        if not hasattr(name, "__hash__"):
            raise InvalidArgument("name of 'Same' object must be hashable")
        self.name = name

        def validate_same(other):
            return self.state_check(other)

        super().__init__(AllOf(Some(*args), validate_same))  # todo: test

    def state_check(self, other):
        if SameState._allow[self.state_name] is False:  # noqa
            raise SameOutsideExpect("Same was used outside of an expect")

        if self.some != other:
            return False

        if self.name not in SameState._state[self.state_name]:
            SameState._state[self.state_name][self.name] = other  # noqa
            return True

        return self._eq(other)

    def _eq(self, other):
        return other == SameState._state[self.state_name][self.name]  # noqa


class NotSame(Same):
    state_name = "NotSame"

    def _eq(self, other):
        return not other == SameState._state[self.state_name][self.name]  # noqa


# alias
is_same = Same
is_unique = NotSame



class default_name:
    def __hash__(self):
        return hash(str(self))


class SameState:
    _allow = {
        "Same": False,
        "NotSame": False
    }
    _state = {
        "Same": {},
        "NotSame": {}
    }

    @staticmethod
    def _start():
        for key in SameState._allow.keys():
            SameState._allow[key] = True
            SameState._state[key] = {}

    @staticmethod
    def _end():
        for key in SameState._allow.keys():
            SameState._allow[key] = False
            SameState._state[key] = {}
